@@ -169,6 +169,7 @@ static unsigned int numsent = 0;            /* packets sent */
 static unsigned int numrecvd = 0;           /* packets received */
 static unsigned int max_replies = UINT_MAX; /* exit after -C replies */
 static unsigned int numdots = 0;            /* dots that should be printed */
+static const char* timestamp_type = NULL;   /* Incoming packet measurement ts type (-m) */
 
 static double stats_min_time = -1;
 static double stats_max_time = -1;
@@ -318,6 +319,103 @@ drop_privileges()
         drop_capabilities();
 }
 
+
+/**
+ * Do pcap_open_live(), except by using the pcap_create() interface
+ * introduced in 2008 (libpcap 0.4) where available.
+ *
+ * FIXME: Use pcap_set_buffer_size()?
+ */
+static pcap_t*
+try_pcap_open_live(const char *device, int snaplen,
+                   int promisc, int to_ms, char *errbuf)
+{
+#ifdef HAVE_PCAP_CREATE
+        pcap_t* pcap;
+        int rc;
+
+        if (!(pcap = pcap_create(device, errbuf))) {
+                goto err;
+        }
+        if ((rc = pcap_set_snaplen(pcap, snaplen))) {
+                snprintf(errbuf, PCAP_ERRBUF_SIZE, "pcap_set_snaplen(): %s", pcap_statustostr(rc));
+                goto err;
+        }
+        if ((rc = pcap_set_promisc(pcap, promisc))) {
+                snprintf(errbuf, PCAP_ERRBUF_SIZE, "pcap_set_promisc(): %s", pcap_statustostr(rc));
+                goto err;
+        }
+        if ((rc = pcap_set_timeout(pcap, to_ms))) {
+                snprintf(errbuf, PCAP_ERRBUF_SIZE, "pcap_set_timeout(): %s", pcap_statustostr(rc));
+                goto err;
+        }
+
+#ifdef HAVE_PCAP_SET_IMMEDIATE_MODE
+        // Without immediate mode some architectures (e.g. Linux with
+        // TPACKET_V3) will buffer replies and incorrectly upwards of
+        // hundreds of milliseconds of delay.
+        if ((rc = pcap_set_immediate_mode(pcap, 1))) {
+                if (verbose) {
+                        fprintf(stderr, "arping: pcap_set_immediate_mode() failed: %s\n", pcap_statustostr(rc));
+                }
+        }
+#endif
+#ifdef HAVE_PCAP_LIST_TSTAMP_TYPES
+        if (timestamp_type) {
+                int err;
+                int v = pcap_tstamp_type_name_to_val(timestamp_type);
+                if (v == PCAP_ERROR) {
+                        fprintf(stderr, "arping: Unknown timestamp type \"%s\"\n", timestamp_type);
+                        exit(1);
+                }
+                if ((err = pcap_set_tstamp_type(pcap, v))) {
+                        fprintf(stderr,
+                                "arping: Failed to set timestamp type \"%s\" (%d): %s\n",
+                                timestamp_type, v, pcap_statustostr(err));
+                }
+        }
+#endif
+        if ((rc = pcap_activate(pcap))) {
+                if (timestamp_type) {
+                        snprintf(errbuf, PCAP_ERRBUF_SIZE, "pcap_activate(tstype=\"%s\"): %s. Try without setting timestamp type.", timestamp_type, pcap_statustostr(rc));
+                } else {
+                        snprintf(errbuf, PCAP_ERRBUF_SIZE, "pcap_activate(): %s", pcap_statustostr(rc));
+                }
+                goto err;
+        }
+#ifdef HAVE_PCAP_LIST_TSTAMP_TYPES
+        // List timestamp types after activating, since we don't want to list
+        // them if activating failed.
+        if (verbose > 1) {
+                int *ts;
+                int count;
+                count = pcap_list_tstamp_types(pcap, &ts);
+                if (count == PCAP_ERROR) {
+                        fprintf(stderr, "arping: pcap_list_tstamp_types() failed\n");
+                } else {
+                        int c;
+                        const char* fmt = "  %-18s %s\n";
+                        fprintf(stderr, "Timestamp types:\n");
+                        fprintf(stderr, fmt, "Name", "Description");
+                        for (c = 0; c < count; c++) {
+                                fprintf(stderr, fmt, pcap_tstamp_type_val_to_name(ts[c]),
+                                        pcap_tstamp_type_val_to_description(ts[c]));
+                        }
+                        pcap_free_tstamp_types(ts);
+                }
+        }
+#endif
+        return pcap;
+err:
+        if (pcap) {
+                pcap_close(pcap);
+        }
+        return NULL;
+#else
+        return pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
+#endif
+}
+
 /**
  * Some stupid OSs (Solaris) think it's a good idea to put network
  * devices in /dev and then play musical chairs with them.
@@ -335,22 +433,22 @@ do_pcap_open_live(const char *device, int snaplen,
         pcap_t* ret;
         char buf[PATH_MAX];
 
-        if ((ret = pcap_open_live(device, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(device, snaplen, promisc, to_ms, errbuf))) {
                 return ret;
         }
 
         snprintf(buf, sizeof(buf), "/dev/%s", device);
-        if ((ret = pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
                 return ret;
         }
 
         snprintf(buf, sizeof(buf), "/dev/net/%s", device);
-        if ((ret = pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
+        if ((ret = try_pcap_open_live(buf, snaplen, promisc, to_ms, errbuf))) {
                 return ret;
         }
 
         /* Call original again to reset the error message. */
-        return pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
+        return try_pcap_open_live(device, snaplen, promisc, to_ms, errbuf);
 }
 
 /**
@@ -490,6 +588,12 @@ extended_usage()
 	       "    -h     Displays a help message and exits.\n"
 	       "    -i interface\n"
 	       "           Use the specified interface.\n"
+               "    -m type"
+#ifndef HAVE_PCAP_LIST_TSTAMP_TYPES
+               " (Disabled on this system. Option ignored)"
+#endif
+               "\n           Type of timestamp to use for incoming packets. Use -vv when\n"
+               "           pinging to list available ones.\n"
 	       "    -q     Does not display messages, except error messages.\n"
 	       "    -r     Raw output: only the MAC/IP address is displayed for each reply.\n"
 	       "    -R     Raw output: Like -r but shows \"the other one\", can  be  combined\n"
@@ -536,7 +640,7 @@ standard_usage()
                "[ -T <host/ip ] "
                "[ -s <MAC> ] [ -t <MAC> ] [ -c <count> ]\n"
                "              "
-               "[ -C <count> ] [ -i <interface> ] "
+               "[ -C <count> ] [ -i <interface> ] [ -m <type> ] "
                "<host/ip/MAC | -B>\n");
 }
 
@@ -1212,7 +1316,7 @@ int main(int argc, char **argv)
 	memcpy(dstmac, ethxmas, ETH_ALEN);
 
         while (EOF != (c = getopt(argc, argv,
-                                  "0aAbBC:c:dDeFhi:I:pPqrRs:S:t:T:uUvw:W:"))) {
+                                  "0aAbBC:c:dDeFhi:I:m:pPqrRs:S:t:T:uUvw:W:"))) {
 		switch(c) {
 		case '0':
 			srcip = 0;
@@ -1265,6 +1369,9 @@ int main(int argc, char **argv)
 		case 'I': /* FALL THROUGH */
 			ifname = optarg;
 			break;
+                case 'm':
+                        timestamp_type = optarg;
+                        break;
 		case 'p':
 			promisc = 1;
 			break;
@@ -1551,6 +1658,7 @@ int main(int argc, char **argv)
 
 #ifdef BIOCIMMEDIATE
 	{
+                // This may be redundant if pcap_set_immediate_mode() is present.
 		uint32_t on = 1;
 		if (0 < (ioctl(pcap_fileno(pcap), BIOCIMMEDIATE,
 			       &on))) {
